/*
** Copyright 2017-2021 Double Precision, Inc.
** See COPYING for distribution information.
*/
#include "libcxxw_config.h"
#include "editor.H"
#include "editor_impl.H"
#include "image.H"
#include "cursor_pointer_element.H"
#include "icon_images_set_element.H"
#include "drag_source_element.H"
#include "drag_destination_element.H"
#include "x/w/tooltip.H"
#include "x/w/gridlayoutmanager.H"
#include "x/w/gridfactory.H"
#include "x/w/input_field.H"
#include "x/w/input_field_appearance.H"
#include "x/w/tooltip_appearance.H"
#include "x/w/input_field_filter.H"
#include "x/w/impl/theme_font_element.H"
#include "x/w/impl/background_color_element.H"
#include "screen.H"
#include "x/w/impl/draw_info.H"
#include "x/w/impl/themedimfwd.H"
#include "busy.H"
#include "defaulttheme.H"
#include "x/w/impl/fonts/current_fontcollection.H"
#include "x/w/impl/fonts/fontcollection.H"
#include "x/w/impl/focus/focusable_element.H"
#include "label_element.H"
#include "x/w/impl/richtext/richtext.H"
#include "richtext/richtext_impl.H"
#include "x/w/richtext/richtextiterator.H"
#include "richtext/richtext_draw_info.H"
#include "selection/current_selection.H"
#include "selection/current_selection_paste_handler.H"
#include "x/w/impl/background_color.H"
#include "messages.H"
#include "connection_thread.H"
#include "generic_window_handler.H"
#include "x/w/impl/icon.H"
#include "catch_exceptions.H"
#include "x/w/key_event.H"
#include "x/w/button_event.H"
#include "x/w/motion_event.H"
#include "popup/popup_impl.H"
#include "popup/popup_handler.H"
#include "popup/popup_attachedto_info.H"
#include <x/vector.H>
#include <x/weakcapture.H>
#include <X11/keysym.h>
#include <courier-unicode.h>
#include <chrono>
#include <iterator>
#include <algorithm>
#include <xcb/xproto.h>

LIBCXXW_NAMESPACE_START

static inline richtextmeta
create_default_meta(const container_impl &container,
		    const input_field_config &config)
{
	auto &element=container->container_element_impl();

	auto bg_color=element.create_background_color
		(config.appearance->foreground_color);

	auto font=element.create_current_fontcollection
		(config.password_char
		 ? config.appearance->password_font
		 : config.appearance->regular_font);

	return {bg_color, font};
}

editor_implObj::init_args
::init_args(const ref<editor_peephole_implObj> &parent_peephole,
	    const text_param &text,
	    bool validated,
	    const input_field_config &config,
	    const input_field_validation_callbackptr &validation_callback,
	    const icon &left_to_right_icon,
	    const icon &right_to_left_icon)
	: parent_peephole{parent_peephole},
	  text{text},
	  validated{validated},
	  config{config},
	  validation_callback{validation_callback},
	  left_to_right_icon{left_to_right_icon},
	  right_to_left_icon{right_to_left_icon},
	  // The theme lock exists before we create the rich text string
	  // that goes into the label implementation superclass
	  theme_lock{parent_peephole->container_element_impl()
		     .get_window_handler().get_screen()
		     ->impl->current_theme},
	  textlabel_config_args{label_config_args},
	  default_meta{create_default_meta(parent_peephole, config)},
	  hint_meta{default_meta} // Fixed up in create_initial_string()
{
	label_config_args.alignment=config.alignment;
	label_config_args.direction=config.direction;

	textlabel_config_args.width_in_columns=config.columns;
	textlabel_config_args.fixed_width_metrics=true;

	if (config.columns < 2)
		throw EXCEPTION("Input fields must have at least two columns");

	if (config.oneline())
		textlabel_config_args.width_in_columns=0;

	textlabel_config_args.child_element_init
		.scratch_buffer_id="textedit@libcxx.com";

	// Initial background_color
	textlabel_config_args.child_element_init
		.background_color=config.appearance->background_color;
}

editor_implObj::init_args::~init_args()=default;

static inline richtextstring
create_initial_string(editor_implObj::init_args &args,
		      richtext_password_info &password_info)
{
	auto &element=args.parent_peephole->container_element_impl();

	auto string=element.create_richtextstring(args.default_meta,
						  args.text);

	if (string.get_meta().size() > 1)
		throw EXCEPTION(_("Input text cannot contain embedded formatting."));

	// If there's a hint, its default font will be the same as the editing
	// font, just need to update its color.
	args.hint_meta=string.get_meta().begin()->second;
	args.hint_meta.textcolor=element.create_background_color
		(args.config.appearance->hint_color);

	// We get called from editor_implObj's constructor.
	// password_info is the superclass, which is already fully
	// constructed.

	if (password_info.password_char != 0)
	{
		// Save the real string here, and replace everything except
		// the appended space with the password_char

		password_info.real_string=string;

		auto cpy=args.text;

		auto b=cpy.string.begin();
		auto e=cpy.string.end();

		std::fill(b, e, password_info.password_char);

		string=element.create_richtextstring(args.default_meta, cpy);
	}

	return string;
}

////////////////////////////////////////////////////////////////////////////

// Avoid explicit conversion to the callback_trigger_t parameter to
// modifying_text's constructor. The trigger object must exist until
// the modifying_text helper object gets destroyed

namespace {
#if 0
}
#endif

struct trigger_parameter {

	const callback_trigger_t &trigger;

	trigger_parameter(const callback_trigger_t &trigger)
		: trigger{trigger}
	{
	}
};

#if 0
{
#endif
}


// Helper object constructed on the stack before modifying the edited text.
//
// The constructor turns off the blinking cursor.
//
// The destructor reenables the blinking cursor.

struct editor_implObj::modifying_text {

protected:
	ONLY IN_THREAD;
	editor_implObj &me;
public:
	selection_cursor_t::lock cursor_lock;
	const input_change_type change_type;
	const callback_trigger_t &trigger;

	modifying_text(ONLY IN_THREAD, editor_implObj &me,
		       input_change_type change_type,
		       const trigger_parameter &trigger)
		: IN_THREAD{IN_THREAD}, me{me},
		  cursor_lock{IN_THREAD, me},
		  change_type{change_type},
		  trigger{trigger.trigger}
	{
		// Make sure blinking is turned off.
		me.unblink(IN_THREAD);
		me.deleted_count=0;
		me.inserted_count=0;
	}

	~modifying_text()
	{
		// Must check this, and call draw_changes() before blink().

		// Changes to the contents update the metadata of what needs
		// to be redrawn, so this must be taken care of first, before
		// letting blink() draw what it wants to draw.

		try {
			if (me.deleted_count || me.inserted_count)
			{
				me.draw_changes(IN_THREAD, cursor_lock,
						change_type,
						me.deleted_count,
						me.inserted_count,
						trigger);
			}

			if (me.current_keyboard_focus(IN_THREAD))
				me.blink(IN_THREAD);
		} CATCH_EXCEPTIONS;
	}
};

namespace {
	// The reason for constructing moving_cursor

	enum class move_type {

		// Normal cursor movement
		navigation,

		// Making a selection
		selection_in_progress,

		// Another application made a selection, so we clear ours.
		clearing_selection,

		// Any other reason not explicitly mentioned here.
		other
	};
};

// Helper object constructed on the stack before moving the cursor, and
// destroyed after the cursor is moved. Inherits from modifying_text, in
// order to turn off/on the blinking cursor, concurrent with the move.
//
// If the SHIFT key or button #1 is held down, and there is no current
// selection, a new selection is started.
//
// When there's a selection, the starting cursor position is saved, and
// the destructor redraws all fragments between the former cursor position and
// the current one, in order to update the display of selected text.
//
// If the SHIFT key is not held down, and there is a current selection, it
// is removed and the formerly selected text fragments get redrawn, to show
// that the selection has been removed.

struct editor_implObj::moving_cursor :
	public modifying_text {

	bool in_selection=false;
	richtextiterator old_cursor;

	// moved is passed to the constructor. The destructor initialized
	// the bool to indicate whether we actually moved.

	bool &moved;
	moving_cursor(ONLY IN_THREAD, editor_implObj &me,
		      const input_mask &mask, bool &moved,
		      const trigger_parameter &trigger)
		: moving_cursor{
				IN_THREAD, me,

				// Either we're making a selection or doing
				// normal navigation, depending on the shift
				// key.
				mask.shift || (mask.buttons & 1)
				? move_type::selection_in_progress
				: move_type::navigation,
				moved, trigger}
	{
	}

	moving_cursor(ONLY IN_THREAD, editor_implObj &me,
		      move_type type,
		      bool &moved,
		      const trigger_parameter &trigger)
		: modifying_text{IN_THREAD, me,
				 // If there's an autocomplete callback, and
				 // it returns something to autocomplete,
				 // set() gets called which recursively called
				 // from draw_changes(), called from
				 // modifying_text's destructor. We depend
				 // on this second turn on the merry-go-round
				 // to skip the autocomplete step, because
				 // the input_change_type is set.
				 //
				 // Otherwise, this is not used, at all.
				 input_change_type::set,
				 trigger},
		  old_cursor{me.cursor->clone()},
		moved{moved}
	{
		if (type != move_type::navigation)
			me.clear_bidi_override(IN_THREAD);

		if (type == move_type::selection_in_progress)
		{
			// Start the selection.

			if (!cursor_lock.cursor)
				cursor_lock.cursor=me.cursor->clone();
		}
		else
		{
			auto p=cursor_lock.cursor;

			if (p)
			{
				// Remove the highlighted selection.

				cursor_lock.cursor=nullptr;

				me.draw_between(IN_THREAD, p, me.cursor);

				// clearing_selection is when clear()ing
				// the current selection in response to
				// someone else installing a new selection.
				//
				// Otherwise, we are removing the selection
				// ourselves, and need to announce it.

				if (type != move_type::clearing_selection)
					me.remove_primary_selection(IN_THREAD);
			}
		}

		if (cursor_lock.cursor)
			in_selection=true;
	}

	~moving_cursor();
};

////////////////////////////////////////////////////////////////////////////

// Implements current_selectionObj for editor's selected/cut text.

class editor_implObj::selectionObj : public current_selectionObj {

	// My editor object. Must be weakly captured to avoid circular refs.

	weakptr<ptr<editor_implObj>> me;

	// Cut text
	std::u32string cut_text;

public:
	// Whether the cut_text contains only US-ASCII

	bool us_ascii_only() const
	{
		for (auto c:cut_text)
			if (c < 0 || c > 127)
				return false;
		return true;
	}

	// Whether the cut_text contains only ISO-8859-1

	bool iso_8859_only() const
	{
		for (auto c:cut_text)
			if (c < 0 || c > 255)
				return false;
		return true;
	}

	selectionObj(xcb_timestamp_t timestamp, const ref<editor_implObj> &me,
		     const richtextiterator &other);

	~selectionObj()=default;

	void clear(ONLY IN_THREAD) override;

	virtual void clear(ONLY IN_THREAD, const ref<editor_implObj> &me)=0;

	ptr<convertedValueObj> convert(ONLY IN_THREAD, xcb_atom_t type)
		override;

	std::vector<xcb_atom_t> supported(ONLY IN_THREAD) override;
};

class editor_implObj::primary_selectionObj : public selectionObj {

public:

	using selectionObj::selectionObj;

	void clear(ONLY IN_THREAD, const ref<editor_implObj> &me) override
	{
		bool ignored;

		callback_trigger_t trigger{cut_copy_paste{}};

		// Leverage moving_cursor for all the heavy lifting.
		//
		// We pretend that we're moving the cursor without
		// a selection in progress, that's all.
		moving_cursor dummy{IN_THREAD, *me,
			move_type::clearing_selection,
			ignored,
			trigger};

		me->current_primary_selection=nullptr;
	}
};

class editor_implObj::secondary_selectionObj : public selectionObj {

public:

	secondary_selectionObj(xcb_timestamp_t timestamp, const
			       ref<editor_implObj> &me,
			       const richtextiterator &other,
			       xcb_atom_t selection)
		: selectionObj{timestamp, me, other},
		  selection{selection}
	{
	}

	const xcb_atom_t selection;

	void clear(ONLY IN_THREAD, const ref<editor_implObj> &me) override
	{
		me->secondary_selection(IN_THREAD)=nullptr;
	}
};

class editor_implObj::xdnd_selectionObj : public selectionObj {

public:
	using selectionObj::selectionObj;

	void clear(ONLY IN_THREAD, const ref<editor_implObj> &me) override
	{
		me->abort_dragging(IN_THREAD);
	}
};

////////////////////////////////////////////////////////////////////////////

editor_implObj::selection_cursor_t::const_lock
::const_lock(editor_implObj &impl)
	: internal_lock{impl.cursor->my_richtext->impl},
	  cursor{impl.selection_cursor.cursor}
{
}

editor_implObj::selection_cursor_t::const_lock::~const_lock()=default;

richtextiteratorptr editor_implObj::selection_cursor_t::const_lock
::cursor_pos() const
{
	return cursor;
}

richtext_draw_info editor_implObj::selection_cursor_t::const_lock
::get_richtext_draw_info(editor_implObj &me) const
{
	return {me, cursor, me.cursor};
}

editor_implObj::selection_cursor_t::lock::lock(ONLY IN_THREAD,
					       editor_implObj &impl,
					       bool blinking_or_clearing)
	: const_lock{impl}
{
	if (!blinking_or_clearing)
		impl.clear_password_peek(IN_THREAD);
}

editor_implObj::selection_cursor_t::lock::~lock()=default;

////////////////////////////////////////////////////////////////////////////

editor_implObj::editor_config::editor_config(const input_field_config &c)
	: columns{c.columns},
	  rows{c.rows},
	  autoselect{c.autoselect},
	  autodeselect{c.autodeselect},
	  maximum_size{c.maximum_size},
	  update_clipboards{c.update_clipboards},
	  directional_format{c.directional_format},
	  direction_tooltip_appearance{
		  c.appearance->direction_tooltip_appearance
	  }
{
}

static richtext_options hint_richtext_options(editor_implObj::init_args &args)
{
	richtext_options options{halign::center};

	options.directional_format=
		args.config.directional_format;
	options.set_bidi(args.label_config_args.direction);

	return options;
}

editor_implObj::editor_implObj(init_args &args)
	: richtext_password_info{args.config},
	  superclass_t{args.config.appearance->background_color,
			  // Background colors
		       args.config.appearance->disabled_background_color,

		       // Icons
		       args.left_to_right_icon,
		       args.right_to_left_icon,

		       // Invisible pointer cursor
		       args.parent_peephole->container_element_impl()
		       .get_window_handler()
		       .create_icon({args.config.appearance
				     ->invisible_pointer})->create_cursor(),

		       // Dragging cursor pointer
		       args.parent_peephole->container_element_impl()
		       .get_window_handler()
		       .create_icon({args.config.appearance
				     ->dragging_pointer})->create_cursor(),

		       // Dragging cursor pointer
		       args.parent_peephole->container_element_impl()
		       .get_window_handler()
		       .create_icon({args.config.appearance
				     ->dragging_nodrop_pointer})
		       ->create_cursor(),
		       std::forward_as_tuple
		       (std::forward_as_tuple(args.config.appearance
					      ->drag_horiz_start,
					      themedimaxis::width),
			std::forward_as_tuple(args.config.appearance
					      ->drag_vert_start,
					      themedimaxis::height)
			),
		       // Capture the string's font.
		       //
		       // We are used by peepholed_fontelementObj, so we
		       // simply use the theme_fontObj mixin for convenience.
		       args.default_meta.getfont(),

		       // label_elementObj::
		       args.parent_peephole,
		       args.textlabel_config_args,
		       args.config.directional_format,
		       *args.theme_lock,
		       create_initial_string(args, *this)},
	  cursor{this->text->end()},
	  dragged_pos{cursor->clone()},
	  parent_peephole{args.parent_peephole},
	  config{args.config},
	  hint{args.config.hint.string.empty() ? richtextptr{}
	       : richtextptr::create
	       (parent_peephole->container_element_impl()
		.create_richtextstring(args.hint_meta,
				       args.config.hint),
		hint_richtext_options(args))},
	  validation_required_thread_only{!args.validated},
	  validation_callback_thread_only{
		  args.validation_callback
	  }
{
	// The first input field in a window gets focus when its shown.

	autofocus=true;

#ifdef EDITOR_CONSTRUCTOR_DEBUG
	EDITOR_CONSTRUCTOR_DEBUG();
#endif
}

editor_implObj::~editor_implObj()=default;

void editor_implObj::initialize(ONLY IN_THREAD)
{
	superclass_t::initialize(IN_THREAD);
	parent_peephole->recalculate(IN_THREAD, *this);
	(void)should_redraw_to_show_hint(IN_THREAD); // Set initial state.
}

void editor_implObj::theme_updated(ONLY IN_THREAD,
				       const const_defaulttheme &new_theme)
{
	superclass_t::theme_updated(IN_THREAD, new_theme);
	parent_peephole->recalculate(IN_THREAD, *this);
}

std::tuple<dim_t, dim_t> editor_implObj::nominal_size(ONLY IN_THREAD) const
{
	auto fc=default_meta.getfont()->fc(IN_THREAD);

	dim_t w=dim_t::truncate(config.columns *
				(dim_t::value_type)
				fc->nominal_width());

	dim_t h=dim_t::truncate(config.rows *
				(dim_t::value_type)
				fc->height());

	auto &hv=*get_horizvert(IN_THREAD);

	if (w < hv.minimum_horiz_override(IN_THREAD))
		w=hv.minimum_horiz_override(IN_THREAD);

	if (h < hv.minimum_vert_override(IN_THREAD))
		h=hv.minimum_vert_override(IN_THREAD);

	return {w, h};
}

void editor_implObj::set_minimum_override(ONLY IN_THREAD,
					      dim_t horiz_override,
					      dim_t vert_override)
{
	superclass_t::set_minimum_override(IN_THREAD, horiz_override,
					   vert_override);

	parent_peephole->recalculate(IN_THREAD, *this);
}

void editor_implObj::rewrap_due_to_updated_position(ONLY IN_THREAD)
{
	initialize_if_needed(IN_THREAD);

	dim_t minimum=parent_peephole->data(IN_THREAD).current_position
		.width;

	text->minimum_width_override(IN_THREAD, minimum);

	if (hint)
		hint->minimum_width_override(IN_THREAD, minimum);

	// text->rewrap(preferred_width);
}

void editor_implObj::keyboard_focus(ONLY IN_THREAD,
					const callback_trigger_t &trigger)
{
	if (!current_keyboard_focus(IN_THREAD))
		(void)validate_modified(IN_THREAD, trigger);

	superclass_t::keyboard_focus(IN_THREAD, trigger);

	auto full_redraw_scheduled=should_redraw_to_show_hint(IN_THREAD);

	if (full_redraw_scheduled)
		schedule_full_redraw(IN_THREAD);

	blink_if_has_focus(IN_THREAD);

#ifdef EDITOR_FOCUS_DEBUG
	EDITOR_FOCUS_DEBUG();
#endif

	bool deselect=false;

	if (config.autoselect)
	{
		if (current_keyboard_focus(IN_THREAD))
		{
			select_all(IN_THREAD, trigger);
			autoselected=true;
		}
		else if (autoselected)
		{
			// If the entire selection was autoselected on
			// focus gain, remove the selection on focus loss.

			deselect=true;
		}
	}
	else if (current_keyboard_focus(IN_THREAD) &&
		 trigger.index() != callback_trigger_button_event)
	{
		// Move to the logical end of the field if focus is not gained
		// because of a button click.

		to_end(IN_THREAD, {}, trigger);
		clear_bidi_override(IN_THREAD); // Can't hurt.
	}

	if (config.autodeselect && !current_keyboard_focus(IN_THREAD))
		deselect=true;

	if (full_redraw_scheduled)
		; // Yea?
	else if (deselect)
	{
		// Leverage the existing moving_cursor logic to clear
		// the current selection.

		bool ignored;

		moving_cursor moving{
			IN_THREAD, *this,
			move_type::other, ignored,
			trigger
		};
	}
}

void editor_implObj::window_focus_change(ONLY IN_THREAD, bool flag)
{
	blink_if_has_focus(IN_THREAD);
	superclass_t::window_focus_change(IN_THREAD, flag);
}

void editor_implObj::blink_if_has_focus(ONLY IN_THREAD)
{
	if (current_keyboard_focus(IN_THREAD) &&
	    get_window_handler().has_focus(IN_THREAD))
	{
		if (!blinking)
			blink(IN_THREAD);
		scroll_cursor_into_view(IN_THREAD);
	}
	else
	{
		unblink(IN_THREAD);
		blinking=nullptr;
	}
}

void editor_implObj::schedule_blink(ONLY IN_THREAD)
{
	blinking=get_screen()->impl->thread->schedule_callback
		(IN_THREAD,
		 std::chrono::milliseconds{500},
		 // Don't create a lambda that owns a strong ref to me.
		 // Use a weak pointer.
		 [me=make_weak_capture(ref<editor_implObj>(this))]
		 (ONLY IN_THREAD)
		 {
			 auto got=me.get();

			 if (got)
			 {
				 auto &[me]=*got;

				 me->blink(IN_THREAD);
			 }
		 });
}

void editor_implObj::unblink(ONLY IN_THREAD)
{
	if (blinkon)
		blink(IN_THREAD);
}

void editor_implObj::blink(ONLY IN_THREAD)
{
	selection_cursor_t::lock cursor_lock{IN_THREAD, *this, true};

	// We actually blink the cursor only when we are not showing a
	// selection.

	if ( (!cursor_lock.cursor) ||
	     cursor_lock.cursor->compare(cursor) == 0)
	{
		blinkon= !blinkon;

		cursor->set_cursor(IN_THREAD, blinkon);
		schedule_blink(IN_THREAD);
	}
	else
	{
		blinking=nullptr;
		return;
	}

	// Tell do_draw() to draw only the vertical slice defined by the
	// cursor's position. Combined with redrawing only the modified fragment
	// this effectively redraws just the cursor.
	auto at=cursor->at(IN_THREAD);
	auto current_position=data(IN_THREAD).current_position;

	current_position.x=at.position.x;
	current_position.width=at.position.width;
	current_position.y=0;
	text->redraw_whatsneeded(IN_THREAD, *this,
				 cursor_lock.get_richtext_draw_info(*this),
				 get_draw_info(IN_THREAD),
				 {current_position});
}

bool editor_implObj::process_key_event(ONLY IN_THREAD, const key_event &ke)
{
	// Hide the pointer by installing the invisible pointer, here.

	// The peephole's report_motion_event() restores the default pointer.

	if (ke.keypress)
		set_cursor_pointer(IN_THREAD,
				   cursor_pointer_1tag<invisible_pointer>
				   ::tagged_cursor_pointer(IN_THREAD));

	if (ke.keypress)
	{
		if (ke.notspecial())
		{
			// No meta/alt/super/hyper held:

			if (process_keypress(IN_THREAD, ke))
			{
				autoselected=false;
				scroll_cursor_into_view(IN_THREAD);
				return true;
			}
		}
		else if (ke.alt)
			switch (ke.keysym) {
			case XK_Left:
			case XK_KP_Left:
				set_bidi_override(IN_THREAD, UNICODE_BIDI_RL,
						  icon_1tag<right_to_left_icon>
						  ::tagged_icon(IN_THREAD),
						  ke);
				return true;
			case XK_Right:
			case XK_KP_Right:
				set_bidi_override(IN_THREAD, UNICODE_BIDI_LR,
						  icon_1tag<left_to_right_icon>
						  ::tagged_icon(IN_THREAD),
						  ke);
				return true;
			}
	}

	if (!ke.keypress)
		switch (ke.keysym) {
		case XK_Shift_L:
		case XK_Shift_R:
			create_primary_selection(IN_THREAD);
		}

	return superclass_t::process_key_event(IN_THREAD, ke);
}

void editor_implObj
::set_bidi_override(ONLY IN_THREAD,
		    unicode_bidi_level_t new_bidi_override,
		    const icon &tooltip_icon,
		    const input_mask &mask)
{
	if (bidi_override == new_bidi_override)
	{
		clear_bidi_override(IN_THREAD); // Toggle.
		return;
	}

	clear_bidi_override(IN_THREAD); // Change of direction.

	bidi_override=new_bidi_override;

	report_cursor_pos_as_motion_event(IN_THREAD);

	direction_attached_popup(IN_THREAD)=show_tooltip
		(IN_THREAD,
		 [&]
		 (ONLY IN_THREAD,
		  const auto &factory)
		 {
			 factory([&]
				 (const container &c)
			 {
				 gridlayoutmanager lm=c->get_layoutmanager();

				 auto f=lm->append_row();
				 f->padding(0);

				 auto image_impl=
					 ref<imageObj::implObj>::create
					 (image_impl_init_params
					  {f->get_container_impl(),
					   tooltip_icon});

				 auto i=image::create(image_impl);

				 f->created_internally(i);
			 },
				 new_gridlayoutmanager{},
				 config.direction_tooltip_appearance);
		 });
}

void editor_implObj::hide_popups(ONLY IN_THREAD)
{
	clear_bidi_override(IN_THREAD);
	superclass_t::hide_popups(IN_THREAD);
}

void editor_implObj::clear_bidi_override(ONLY IN_THREAD)
{
	bidi_override=UNICODE_BIDI_SKIP;

	direction_attached_popup(IN_THREAD)=nullptr;
}

void editor_implObj::update_attachedto_info(ONLY IN_THREAD)
{
	if (bidi_override != UNICODE_BIDI_SKIP &&
	    direction_attached_popup(IN_THREAD))
	{
		const auto &[pos, ignore] = tooltip_pseudo_rectangle
			(IN_THREAD,
			 config.direction_tooltip_appearance);

		direction_attached_popup(IN_THREAD)->impl->handler
			->update_attachedto_element_position
			(IN_THREAD, pos);
		return;
	}
	superclass_t::update_attachedto_info(IN_THREAD);
}

void editor_implObj::report_cursor_pos_as_motion_event(ONLY IN_THREAD)
{
	// Simulate a motion event so that the tooltip comes up where the
	// cursor is.

	auto pos=cursor->at(IN_THREAD).position;
	input_mask mask;

	motion_event me{mask, motion_event_type::keyboard_action_event,
		coord_t::truncate(pos.x + pos.width/2),
		coord_t::truncate(pos.y),
	};
	report_motion_event(IN_THREAD, me);
}

bidi editor_implObj::direction(ONLY IN_THREAD)
{
	switch (bidi_override) {
	case UNICODE_BIDI_LR:
		return bidi::left_to_right;
	case UNICODE_BIDI_RL:
		return bidi::right_to_left;
	default:
		break;
	}

	return superclass_t::direction(IN_THREAD);
}

void editor_implObj::set_cursor_pointer(ONLY IN_THREAD,
					    const cursor_pointer &p)
{
	parent_peephole->get_element_impl().set_cursor_pointer(IN_THREAD, p);
}

xcb_atom_t editor_implObj::secondary_clipboard(ONLY IN_THREAD)
{
	return IN_THREAD->info->get_atom
		(get_window_handler().get_screen()->impl->current_theme.get()
		 ->default_cut_paste_selection());
}

bool editor_implObj::process_keypress(ONLY IN_THREAD, const key_event &ke)
{
	if (ke.ctrl)
	{
		if (ke.keysym == XK_Insert)
		{
			create_secondary_selection
				(IN_THREAD,
				 secondary_clipboard(IN_THREAD));
			return true;
		}

		if (ke.unicode == 1) // CTRL-A
		{
			select_all(IN_THREAD, &ke);
			return true;
		}

		if (ke.unicode == ' ')
		{
			get_window_handler()
				.receive_selection(IN_THREAD, XCB_ATOM_PRIMARY);
			return true;
		}

		return false;
	}

	bool moved;

	switch (ke.keysym) {
	case XK_Left:
	case XK_KP_Left:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->left(IN_THREAD);
		}
		return moved;
	case XK_Right:
	case XK_KP_Right:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->right(IN_THREAD);
		}
		return moved;
	case XK_Up:
	case XK_KP_Up:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->up(IN_THREAD);
		}
		return moved;
	case XK_Down:
	case XK_KP_Down:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->down(IN_THREAD);
		}
		return moved;
	case XK_Delete:
	case XK_KP_Delete:
		delete_char_or_selection(IN_THREAD, ke, &ke);
		return true;
	case XK_Page_Up:
	case XK_KP_Page_Up:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->page_up(IN_THREAD,
					bounds(get_draw_info(IN_THREAD)
					       .element_viewport).height);
		}
		return moved;
	case XK_Page_Down:
	case XK_KP_Page_Down:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->page_down(IN_THREAD,
					  bounds(get_draw_info(IN_THREAD)
						 .element_viewport).height);
		}
		return moved;
	case XK_Home:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->start_of_line();
		}
		return moved;
	case XK_End:
		{
			callback_trigger_t trigger{&ke};

			moving_cursor moving{IN_THREAD, *this, ke, moved,
					     trigger};
			cursor->end_of_line();
		}
		return moved;
	case XK_KP_Home:
		return to_begin(IN_THREAD, ke, &ke);
	case XK_KP_End:
		return to_end(IN_THREAD, ke, &ke);
	case XK_Insert:
	case XK_KP_Insert:
		if (ke.shift)
			get_window_handler().receive_selection
				(IN_THREAD,
				 secondary_clipboard(IN_THREAD));
		return true;
	}

	if ((!config.oneline() && ke.unicode == '\n') || ke.unicode >= ' ')
	{
		char32_t string[4];

		switch (bidi_override) {
		case UNICODE_BIDI_LR:
			string[0]=UNICODE_LRM;
			string[1]=UNICODE_LRO;
			string[2]=ke.unicode;
			string[3]=UNICODE_LRM;
			insert(IN_THREAD, {string, 4}, &ke);
			break;
		case UNICODE_BIDI_RL:
			string[0]=UNICODE_RLM;
			string[1]=UNICODE_RLO;
			string[2]=ke.unicode;
			string[3]=UNICODE_RLM;
			insert(IN_THREAD, {string, 4}, &ke);
			break;
		default:
			insert(IN_THREAD, {&ke.unicode, 1}, &ke);
			break;
		}
		return true;
	}

	if (ke.unicode == '\n')
	{
		// Must be an Enter in a single-line field. We intuitively
		// expect to validate the input field's contents, if nothing
		// else happens.
		if (!validate_modified(IN_THREAD, &ke))
			return true; // We did something

		// Otherwise, fall through, and we didn't do anything.
	}

	if (ke.unicode == '\b')
	{
		callback_trigger_t trigger{&ke};

		modifying_text modifying{IN_THREAD, *this,
					 input_change_type::deleted,
					 trigger};
		delete_selection_info del_info{IN_THREAD, *this, modifying};

		// For correct cursor behavior in right to left text,
		// we move to the left manually, first.

		auto delete_cursor=cursor->clone();
		if (delete_cursor->left(IN_THREAD))
		{
			delete_char(IN_THREAD, modifying, delete_cursor);
		}
		return true;
	}
	return false;
}

inline
input_field_filter_info::input_field_filter_info(input_filter_type type,
						 bidi direction,
						 const richtextiterator &starting_pos,
						 const richtextiterator &ending_pos,
						 const std::u32string_view
						 &new_contents,
						 size_t size,
						 size_t maximum_size)
	: type{type}, direction{direction},
	  starting_pos{starting_pos}, ending_pos{ending_pos},
	  n_deleted{starting_pos->count_richtextstring(ending_pos)},
	  new_contents{new_contents}, size{size}, maximum_size{maximum_size}
{
}

void input_field_filter_info::update() const
{
	update(starting_pos, ending_pos, new_contents);
}

struct editor_implObj::input_field_filter_info_impl
	: public input_field_filter_info {

	ONLY IN_THREAD;
	editor_implObj &me;
	modifying_text &modifying;

	//! Constructor
	inline input_field_filter_info_impl(ONLY IN_THREAD,
					    editor_implObj &me,
					    input_filter_type type,
					    const richtextiterator
					    &starting_pos,
					    const richtextiterator &ending_pos,
					    const std::u32string_view
					    &new_contents,
					    modifying_text &modifying)
		: input_field_filter_info
		{type,
		 (*modifying.cursor_lock.internal_lock)->rl()
		 ? bidi::right_to_left : bidi::left_to_right,
		 starting_pos, ending_pos,
		 new_contents, me.size(),
		 me.config.maximum_size
		},
		  IN_THREAD{IN_THREAD},
		  me{me},
		  modifying{modifying}
	{
	}

	void update(const richtextiterator &starting_pos,
		    const richtextiterator &ending_pos,
		    const std::u32string_view &new_contents) const override
	{
		me.update_filtered_content(IN_THREAD, starting_pos,
					   ending_pos,
					   new_contents);
	}

	void move(size_t pos) const override
	{
		me.cursor->swap(me.cursor->pos(pos));
	}

	void select_all() const override
	{
		me.cursor->swap(me.cursor->end());
		modifying.cursor_lock.cursor=me.cursor->begin();
		me.schedule_full_redraw(IN_THREAD);
	}
};

// Implement original_pos(), for an input_field_filter_info, for modification.
// original_pos() returns same value as starting_pos().

struct editor_implObj::input_field_filter_info_impl_change
	: public input_field_filter_info_impl {

	inline input_field_filter_info_impl_change
	(ONLY IN_THREAD,
	 editor_implObj &me,
	 modifying_text &modifying,
	 const richtextiterator &a,
	 const richtextiterator &b,
	 const std::u32string_view &new_contents)
		: input_field_filter_info_impl
		{
		 IN_THREAD,
		 me,
		 modifying.change_type==input_change_type::inserted
		 ? input_filter_type::inserting
		 : modifying.change_type==input_change_type::deleted
		 ? input_filter_type::deleting
		 : input_filter_type::replacing,
		 a, b,
		 new_contents,
		 modifying
		}
	{
	}

	size_t original_pos() const override
	{
		return starting_pos->pos();
	}
};

// Implement original_pos(), for an input_field_filter_info, for cursor
// movement.
//
// original_pos() retrieves the original cursor's pos().

struct editor_implObj::input_field_filter_info_impl_move
	: public input_field_filter_info_impl {

	const richtextiterator &original;

	inline input_field_filter_info_impl_move(ONLY IN_THREAD,
						 editor_implObj &me,
						 const richtextiterator
						 &original,
						 modifying_text &modifying)
		: input_field_filter_info_impl
		{
		 IN_THREAD,
		 me,
		 input_filter_type::move_only,
		 me.cursor,
		 me.cursor,
		 U"",
		 modifying
		},
		  original{original}
	{
	}

	size_t original_pos() const override
	{
		return original->pos();
	}
};

editor_implObj::moving_cursor::~moving_cursor()
{
	moved=old_cursor->compare(me.cursor) != 0;

	if (moved && me.on_filter(IN_THREAD))
	{
		input_field_filter_info_impl_move impl{IN_THREAD,
						       me,
						       old_cursor,
						       *this};

		try {
			me.on_filter(IN_THREAD)(IN_THREAD, impl);
		} REPORT_EXCEPTIONS((&me));
	}

	if (in_selection)
	{
		me.draw_between(IN_THREAD,
				old_cursor,
				me.cursor);
	}
}

void editor_implObj
::update_content(ONLY IN_THREAD,
		 modifying_text &modifying,
		 const current_selection_info &info,
		 const std::u32string_view &str)
{
	auto starting_position=info.starting_position;
	auto ending_position=info.ending_position;

	if (ending_position->pos() < starting_position->pos())
		std::swap(starting_position, ending_position);

	if (on_filter(IN_THREAD))
	{
		input_field_filter_info_impl_change
			impl{IN_THREAD,
			*this,
			modifying,
			starting_position,
			ending_position,
			str};

		try {
			on_filter(IN_THREAD)(IN_THREAD, impl);
		} REPORT_EXCEPTIONS(this);
		return;
	}
	update_filtered_content(IN_THREAD, starting_position,
				ending_position, str);
}

void editor_implObj
::update_filtered_content(ONLY IN_THREAD,
			  const richtextiterator &starting_position,
			  const richtextiterator &ending_position,
			  const std::u32string_view &str)
{
	if (!password_char)
	{
		// Straightforward, use unicode_bidi_cleaned_size() to
		// quickly compute the actual character count without
		// bothering to modify the string.

		update_filtered_content(IN_THREAD, starting_position,
					ending_position,
					str,
					unicode_bidi_cleaned_size
					(str.data(),
					 str.size(),
					 UNICODE_BIDI_CLEANUP_EXTRA));
		return;
	}

	// For passwords we will strip out all directional and isolation marks,
	// since we are masking the text anyway.
	std::u32string cleaned_str{str.begin(), str.end()};

	unicode::bidi_cleanup(cleaned_str, [](size_t){},
			      UNICODE_BIDI_CLEANUP_EXTRA);

	update_filtered_content(IN_THREAD, starting_position,
				ending_position,
				cleaned_str,
				cleaned_str.size());
}

void editor_implObj
::update_filtered_content(ONLY IN_THREAD,
			  richtextiterator starting_position,
			  richtextiterator ending_position,
			  const std::u32string_view &str,
			  size_t cleaned_size)
{
	if (ending_position->pos() < starting_position->pos())
		std::swap(starting_position, ending_position);

	size_t n=starting_position->count_richtextstring(ending_position);

	if (n)
	{
		if (size() - n + cleaned_size > config.maximum_size)
			return;

		deleted_count += n;

		if (n)
		{
			starting_position->remove(IN_THREAD, ending_position);

			if (password_char)
			{
				// Extra work for password fields.

				mpobj<richtextstring>::lock lock{real_string};

				lock->erase(starting_position->pos(), n);
			}
		}
	}
	else
	{
		if (size() + cleaned_size > config.maximum_size)
			return;
	}
	inserted_count += cleaned_size;

	if (password_char == 0)
	{
		starting_position->insert(IN_THREAD, str);
		return;
	}

	// Extra work for password fields.

	if (str.size() == 1 &&
	    starting_position->end()->compare(starting_position) == 0)
	{
		password_peeking=get_screen()->impl->thread->schedule_callback
			(IN_THREAD,
			 std::chrono::seconds{2},
			 // Don't create a lambda that owns a strong ref to me.
			 // Use a weak pointer.
			 [me=make_weak_capture(ref(this))]
			 (ONLY IN_THREAD)
			 {
				 auto got=me.get();

				 if (got)
				 {
					 auto & [me]=*got;
					 me->clear_password_peek(IN_THREAD);
				 }
			 });

		// TODO: support bidi here, somehow.
		char32_t left_right_text[]={
			UNICODE_LRM,
			UNICODE_LRO,
			str[0],
			UNICODE_LRM,
			0
		};
		starting_position->insert(IN_THREAD, left_right_text);
	}
	else
	{
		starting_position->insert(IN_THREAD,
					  std::u32string(str.size(),
							 password_char));
	}

	mpobj<richtextstring>::lock lock{real_string};

	lock->insert(starting_position->pos(), str);
}

void editor_implObj::clear_password_peek(ONLY IN_THREAD)
{
	if (!password_peeking)
		return;

	selection_cursor_t::lock cursor_lock{IN_THREAD, *this, true};

	password_peeking=nullptr;

	auto p=cursor->clone();

	p->move(IN_THREAD, -1);
	p->insert(IN_THREAD, std::u32string(1, password_char));
	p->remove(IN_THREAD, cursor);

	text->redraw_whatsneeded(IN_THREAD, *this,
				 cursor_lock.get_richtext_draw_info(*this),
				 get_draw_info(IN_THREAD));
}

richtextstring editor_implObj::get_content(const richtextiterator &a,
					       const richtextiterator &b,
					       const std::optional<bidi_format>
					       &embedding)
{
	if (password_char == 0)
		return a->get_richtextstring(b, embedding);

	// Look at the real_string, instead.

	auto ap=a->pos();
	auto bp=b->pos();

	if (ap > bp)
		std::swap(ap, bp);

	mpobj<richtextstring>::lock lock{real_string};

	if (lock->size() == 0)
		return *lock;

	size_t l=bp-ap;

	if (ap >= lock->size())
		ap=lock->size()-1;

	if (l > lock->size()-ap)
		l=lock->size()-ap;

	return {*lock, ap, l};
}

bool editor_implObj::uses_input_method()
{
	return true;
}

bool editor_implObj::pasted(ONLY IN_THREAD,
				const std::u32string_view &str)
{
	callback_trigger_t trigger{cut_copy_paste{}};

	insert(IN_THREAD, str, trigger);
	scroll_cursor_into_view(IN_THREAD);
	return true;
}

void editor_implObj::insert(ONLY IN_THREAD,
				const std::u32string_view &str,
				const callback_trigger_t &trigger)
{
	if (str.empty())
		return;

	modifying_text modifying{IN_THREAD, *this,
				 input_change_type::inserted, trigger};
	delete_selection_info del_info{IN_THREAD, *this, modifying};

	update_content(IN_THREAD, modifying, del_info,
		       str);
}

void editor_implObj::enablability_changed(ONLY IN_THREAD)
{
	set_background_color
		(IN_THREAD,
		 enabled(IN_THREAD, enabled_for::input_focus)
		 ? background_color_element<textedit_background_color>
		 ::get(IN_THREAD)
		 : background_color_element<textedit_disabled_background_color>
		 ::get(IN_THREAD));
}

void editor_implObj::draw_changes(ONLY IN_THREAD,
				      selection_cursor_t::lock &cursor_lock,
				      input_change_type change_made,
				      size_t deleted,
				      size_t inserted,
				      const callback_trigger_t &trigger)
{
	recalculate(IN_THREAD,

		    // We handle text direction changes ourselves. We expect
		    // our size to be fixed, and when the default text
		    // direction changes we will redraw the entire text
		    // label with new alignment.

		    false);

	// If we originally should've shown the hint, but now not, or vice
	// versa, we should redraw everything fully.

	if (should_redraw_to_show_hint(IN_THREAD))
		schedule_full_redraw(IN_THREAD);

	// We can be called from set() before the editor element is visible.
	// Don't bother calling get_draw_info(), because that might get
	// stale anyway.
	else if (data(IN_THREAD).logical_inherited_visibility)
		text->redraw_whatsneeded(IN_THREAD, *this,
					 cursor_lock.get_richtext_draw_info
					 (*this),
					 get_draw_info(IN_THREAD));

	validation_required(IN_THREAD)=true;

	auto &cb=on_change(IN_THREAD);

	if (cb)
	{
		try {
			cb(IN_THREAD, input_change_info_t
			   {change_made, inserted, deleted, size(), trigger});
		} REPORT_EXCEPTIONS(this);
	}

	// Invoke the autocomplete callback if the conditions are right.

	size_t s=size();

	bool is_delete_or_backspace=false;

	if (std::holds_alternative<const key_event *>(trigger))
	{
		auto &ke=std::get<const key_event *>(trigger);

		if (ke && (ke->keysym == XK_Delete ||
			   ke->keysym == XK_KP_Delete ||
			   ke->unicode == '\b'))
			is_delete_or_backspace=true;
	}

	if (change_made != input_change_type::set &&
	    s > 0 && cursor->pos() == s &&
	    !is_delete_or_backspace &&
	    on_autocomplete(IN_THREAD) &&
	    (!cursor_lock.cursor || cursor_lock.cursor->pos() == cursor->pos()))
	{
		busy_impl mcguffin{*this};

		input_autocomplete_info_t info{get(std::nullopt), 0, mcguffin};

		bool flag=false;

		auto &cb=on_autocomplete(IN_THREAD);

		try {
			flag=cb(IN_THREAD, info);
		} REPORT_EXCEPTIONS(this);

		if (flag)
			set(IN_THREAD, info.string, info.string.size(),
			    info.selection_start, trigger);
	}
}

void editor_implObj::do_draw(ONLY IN_THREAD,
				 const draw_info &di,
				 const rectarea &areas)
{
#ifdef EDITOR_DRAW
	EDITOR_DRAW();
#endif

	if (show_hint(IN_THREAD))
	{
		hint->full_redraw(IN_THREAD, *this, {*this}, di, areas);
		return;
	}

	selection_cursor_t::lock cursor_lock{IN_THREAD, *this, true};

	text->full_redraw(IN_THREAD, *this,
			  cursor_lock.get_richtext_draw_info(*this),
			  di, areas);
}

void editor_implObj::draw_between(ONLY IN_THREAD,
				      const richtextiterator &a,
				      const richtextiterator &b)
{
	selection_cursor_t::lock cursor_lock{IN_THREAD, *this};

	text->redraw_between(IN_THREAD, *this,
			     a, b,
			     cursor_lock.get_richtext_draw_info(*this),
			     get_draw_info(IN_THREAD));
}

void editor_implObj
::set_focus_and_ensure_visibility(ONLY IN_THREAD,
				  const callback_trigger_t &trigger)
{
	set_focus_only(IN_THREAD, trigger);
}

void editor_implObj::scroll_cursor_into_view(ONLY IN_THREAD)
{
	auto pos=cursor->at(IN_THREAD).position;

	ensure_visibility(IN_THREAD, pos);
	report_current_cursor_position(IN_THREAD, pos);

	if (bidi_override != UNICODE_BIDI_SKIP)
	{
		// We must be showing the bidi tooltip popup, so first
		// update the cursor position as a fake motion event

		report_cursor_pos_as_motion_event(IN_THREAD);

		// And then update the bidi popup's position

		update_attachedto_info(IN_THREAD);
	}
}

bool editor_implObj::process_button_event(ONLY IN_THREAD,
					      const button_event &be,
					      xcb_timestamp_t timestamp)
{
	if (!enabled(IN_THREAD, enabled_for::input_focus))
		return superclass_t::process_button_event(IN_THREAD, be,
							  timestamp);

	autoselected=false;

	if (be.press && (be.button == 1 || be.button == 2

			 // Respond to right mouse button click only if
			 // we do not have input focus already. Otherwise
			 // we might have something already selected, and
			 // don't really want to deselect it.
			 || (be.button == 3 &&
			     !current_keyboard_focus(IN_THREAD)))
	    && be.click_count == 1)
	{
		// First button click.

		bool ignored;

		abort_dragging(IN_THREAD); // Something stale?

		if (be.button == 1)
		{
			selection_cursor_t::lock
				cursor_lock{IN_THREAD, *this};

			// If we're clicking in a middle of a selection,
			// we might end up dragging the selected text.

			if (cursor_lock.cursor)
			{
				richtextiterator a{cursor_lock.cursor};
				richtextiterator b=cursor;

				if (a->compare(b) > 0)
					std::swap(a, b);

				richtextiterator c=b->clone();

				if (c->moveto(IN_THREAD,
					      most_recent_x,
					      most_recent_y) &&
				    a->compare(c) <= 0 &&
				    b->compare(c) > 0)
				{
					auto s=ref<xdnd_selectionObj>::create
						(timestamp, ref(this),
						 cursor_lock.cursor);


					std::vector<xcb_atom_t> source_formats
						{IN_THREAD->info->atoms_info
						 .text_plain_utf8_mime};

					if (s->iso_8859_only())
					{
						source_formats.push_back
							(IN_THREAD->info
							 ->atoms_info
							 .text_plain_mime);
					}

					start_dragging(IN_THREAD,
						       s, source_formats,
						       most_recent_x,
						       most_recent_y);
				}
			}
		}

		// We grab the pointer while the button is held down.
		grab(IN_THREAD, &be);

		// If we figured out we're not dragging currently selected text,
		// we'll do what we normally do when a button is pressed.

		if (!grab_inprogress(IN_THREAD))
		{
			callback_trigger_t trigger{&be};

			moving_cursor moving{IN_THREAD, *this, be, ignored,
					     trigger};

			cursor->moveto(IN_THREAD, most_recent_x, most_recent_y);

			if (be.button == 2)
			{
				set_focus_only(IN_THREAD, &be);
				get_window_handler()
					.receive_selection(IN_THREAD,
							   XCB_ATOM_PRIMARY);
			}
		}
	}
	else if (be.press)
	{
		// Some other button click.

		if (be.button == 1 && be.click_count == 2)
		{
			callback_trigger_t trigger(&be);

			auto [start, end]=cursor->select_word();

			if (start->compare(end))
			{
				bool ignored;

				{
					moving_cursor moving{IN_THREAD, *this,
						move_type::navigation,
						ignored, trigger};

					cursor->swap(start);
				}
				moving_cursor moving{IN_THREAD, *this,
					move_type::selection_in_progress,
					ignored, trigger};
				cursor->swap(end);
			}
		}
	}
	else if (be.button == 1)
	{
		if (grab_inprogress(IN_THREAD))
		{
			release_dragged_selection(IN_THREAD);
		}
		else
		{
			stop_scrolling(IN_THREAD);
			create_primary_selection(IN_THREAD);
		}
	}

	if (!be.press)
		abort_dragging(IN_THREAD);

	// We do not consume the button event. The peephole with this editor
	// element also consumes the button press, and moves the input focus
	// here, see editor_peephole_implObj.

	return superclass_t::process_button_event(IN_THREAD, be, timestamp);
}

void editor_implObj::report_motion_event(ONLY IN_THREAD,
					     const motion_event &me)
{
	most_recent_x=me.x;
	most_recent_y=me.y;

	if (grab_inprogress(IN_THREAD))
	{
		report_dragged_motion_event(IN_THREAD, me);
		return;
	}

	// If we're hiding the pointer, remove it. This is done
	// in editor_peephole_impl:
	superclass_t::report_motion_event(IN_THREAD, me);

	if ((me.mask.buttons & 1) && me.type == motion_event_type::real_motion)
	{
		bool ignored;
		{
			callback_trigger_t trigger{&me};

			moving_cursor moving{IN_THREAD, *this, me.mask,
					     ignored, trigger};
			cursor->moveto(IN_THREAD, most_recent_x, most_recent_y);
		}

		if (!motion_scroll_callback)
			start_scrolling(IN_THREAD);
	}
}

bool editor_implObj::accepts_drop(ONLY IN_THREAD,
				      const source_dnd_formats_t
				      &source_formats,
				      xcb_timestamp_t timestamp)
{
	if (source_formats.count(IN_THREAD->info
				 ->atoms_info.text_plain_utf8_mime) > 0)
	{
		dropped_atom=IN_THREAD->info->atoms_info.text_plain_utf8_mime;
		return true;
	}

	if (source_formats.count(IN_THREAD->info
				 ->atoms_info.text_plain_mime) > 0)
	{
		dropped_atom=IN_THREAD->info->atoms_info.text_plain_mime;
		return true;
	}

	if (source_formats.count(IN_THREAD->info
				 ->atoms_info.text_plain_iso8859_mime) > 0)
	{
		dropped_atom=
			IN_THREAD->info->atoms_info.text_plain_iso8859_mime;
		return true;
	}
	return false;
}

void editor_implObj::dragging_location(ONLY IN_THREAD, coord_t x, coord_t y,
					   xcb_timestamp_t timestamp)
{
	dragged_pos->moveto(IN_THREAD, x, y);
}

current_selection_handlerptr
editor_implObj::drop(ONLY IN_THREAD,
			 xcb_atom_t &type,
			 const ref<obj> &finish_mcguffin)
{
	// The first step is to move the cursor to where the drag was dropped.
	//
	// We're borrowing the cut-paste framework, which pastes text to
	// where the current input focus is.

	bool moved=false;
	{
		callback_trigger_t trigger{cut_copy_paste{}};

		moving_cursor moving{
			IN_THREAD, *this,
			move_type::other,
			moved,
			trigger
		};

		cursor->swap(dragged_pos);
		set_focus_only(IN_THREAD, {});
	}

	type=dropped_atom;

	return current_selection_paste_handler
		::create(std::vector<xcb_atom_t>{}, finish_mcguffin);
}

void editor_implObj::start_scrolling(ONLY IN_THREAD)
{
	motion_scroll_callback=get_screen()->impl->thread->schedule_callback
		(IN_THREAD,
		 std::chrono::milliseconds{250},
		 // Don't create a lambda that owns a strong ref to me.
		 // Use a weak pointer.
		 [me=make_weak_capture(ref(this))]
		 (ONLY IN_THREAD)
		 {
			 auto got=me.get();

			 if (got)
			 {
				 auto & [me]=*got;
				 me->scroll(IN_THREAD);
			 }
		 });
}

void editor_implObj::scroll(ONLY IN_THREAD)
{
	stop_scrolling(IN_THREAD);
	// In order to scroll_cursor_into_view, that's it.
	//
	// If motion_event()s continue, the next one will start_scrolling()
	// again.
}

void editor_implObj::stop_scrolling(ONLY IN_THREAD)
{
	motion_scroll_callback=nullptr;
	scroll_cursor_into_view(IN_THREAD);
}

void editor_implObj::removed(ONLY IN_THREAD)
{
	// When this element is removed from its container, remove all the
	// selections, too...

	remove_primary_selection(IN_THREAD);
	remove_secondary_selection(IN_THREAD);
}

/////////////////////////////////////////////////////////////////////////////
//
// Selections

editor_implObj::selectionObj::
selectionObj(xcb_timestamp_t timestamp, const ref<editor_implObj> &me,
	     const richtextiterator &other)
	: current_selectionObj(timestamp), me(me),
	  cut_text{me->get_content(me->cursor,
				   other,
				   me->get_screen()->impl
				   ->default_bidi_format.get()).get_string()}
{
}

void editor_implObj::selectionObj::clear(ONLY IN_THREAD)
{
	auto p=me.getptr();

	if (!p)
		return;
	clear(IN_THREAD, p);
}

std::vector<xcb_atom_t> editor_implObj::selectionObj
::supported(ONLY IN_THREAD)
{
	std::vector<xcb_atom_t> v;

	auto p=me.getptr();

	if (p)
	{
		const auto &atoms=IN_THREAD->info->atoms_info;

		v.push_back(atoms.text_plain_utf8_mime);
		v.push_back(atoms.utf8_string);

		if (us_ascii_only())
			v.push_back(atoms.string);
		if (iso_8859_only())
		{
			v.push_back(atoms.text_plain_mime);
			v.push_back(atoms.text_plain_iso8859_mime);
		}
	}

	return v;
}

// Somebody is asking for our selection.

ptr<current_selectionObj::convertedValueObj> editor_implObj::selectionObj
::convert(ONLY IN_THREAD, xcb_atom_t type)
{
	const char *charset=nullptr;

	auto p=me.getptr();

	if (p)
	{
		const auto &atoms=p->get_screen()->impl->thread
			->info->atoms_info;

		if (type == atoms.string)
		{
			if (us_ascii_only())
				charset=unicode::iso_8859_1;
		}

		if (type == atoms.text_plain_mime ||
		    type == atoms.text_plain_iso8859_mime)
		{
			if (iso_8859_only())
				charset=unicode::iso_8859_1;
		}
		if (type == atoms.utf8_string ||
		    type == atoms.text_plain_utf8_mime)
			charset=unicode::utf_8;
	}

	if (!charset)
		return ptr<convertedValueObj>();

	auto bytes=vector<uint8_t>::create();

	bytes->reserve(cut_text.size()*2);

	bool ignore;

	unicode::iconvert::fromu::convert(cut_text.begin(),
					  cut_text.end(),
					  charset,
					  std::back_insert_iterator<std::vector
					  <uint8_t>>(*bytes), ignore);

	return ref<current_selectionObj::convertedValueObj>
		::create(type, 8, bytes);
}

/////////////////////////////////////////////////////////////////////////////
//
// Constructing a delete_selection_info. Invoke the delegated constructor
// with the parameters specifying the boundaries of the current selection.
//
// If there's a current selection, the delegated constructor receives
// the pos() of both rich text iterators, otherwise it receives the current
// cursor's position.
//
// The delegated constructor figures out which one of the rich text iterators
// is the first one, and sets the starting_pos and n accordingly.
//
// In this manner, if there is a selection this ends up using whichever iterator
// is earlier as the starting position, and the computed character count;
// and if there is no selection the starting_pos is the current cursor position.

editor_implObj::delete_selection_info
::delete_selection_info(ONLY IN_THREAD,
			editor_implObj &me,
			modifying_text &modifying)
	: delete_selection_info{IN_THREAD, me, modifying, me.cursor,
				modifying.cursor_lock.cursor ?
				richtextiterator{modifying.cursor_lock.cursor}
				: me.cursor}
{
}

editor_implObj::delete_selection_info
::delete_selection_info(ONLY IN_THREAD,
			editor_implObj &me,
			modifying_text &modifying,
			const richtextiterator &starting_position,
			const richtextiterator &ending_position)
	: current_selection_info{starting_position, ending_position},
	  IN_THREAD{IN_THREAD},
	  me{me},
	  cursor_lock{modifying.cursor_lock},
	  had_selection{starting_position->pos() != ending_position->pos()}
{
}

editor_implObj::delete_selection_info::~delete_selection_info()
{
	// Now that we've computed what the current situation is, we
	// can go ahead and remove the primary selection.

	if (had_selection)
	{
		cursor_lock.cursor=richtextiteratorptr();
		me.remove_primary_selection(IN_THREAD);
	}
}

bool editor_implObj::selection_can_be_received()
{
	return true;
}

void editor_implObj::create_primary_selection(ONLY IN_THREAD)
{
	if (!config.update_clipboards)
		return;

	selection_cursor_t::lock cursor_lock{IN_THREAD, *this};

	if (!cursor_lock.cursor)
		return;

	if (cursor_lock.cursor->compare(cursor) == 0)
		return; // Nothing selected, really.
	// Remove the previous one.

	remove_primary_selection(IN_THREAD);

	auto s=ref<primary_selectionObj>
		::create(get_screen()->impl->thread->timestamp(IN_THREAD),
			 ref(this), cursor_lock.cursor);

	current_primary_selection=s;
	get_window_handler().selection_announce(IN_THREAD, XCB_ATOM_PRIMARY, s);
}

bool editor_implObj::cut_or_copy_selection(cut_or_copy_op op,
					       xcb_atom_t selection)
{
	if (op == cut_or_copy_op::available)
		return current_primary_selection.get() ? true:false;

	get_window_handler().thread()
		->run_as([me=ref{this}, op, selection]
			 (ONLY IN_THREAD)
			 {
				 me->cut_or_copy_selection(IN_THREAD,
							   op,
							   selection);
			 });
	return true;
}

bool editor_implObj::cut_or_copy_selection(ONLY IN_THREAD,
					       cut_or_copy_op op,
					       xcb_atom_t selection)
{
	switch (op) {
	case cut_or_copy_op::available:
		return cut_or_copy_selection(op, selection);
	case cut_or_copy_op::copy:
		create_secondary_selection(IN_THREAD, selection);
		break;
	case cut_or_copy_op::cut:
		{
			callback_trigger_t trigger{cut_copy_paste{}};

			modifying_text modifying{IN_THREAD, *this,
						 input_change_type::deleted,
						 trigger};

			delete_selection_info del_info{IN_THREAD, *this,
						       modifying};

			if (create_secondary_selection(IN_THREAD,
						       selection,
						       del_info.cursor_lock))
			{
				update_content(IN_THREAD,
					       modifying,
					       del_info,
					       U"");
			}
		}
		break;
	}
	return true;
}

void editor_implObj::create_secondary_selection(ONLY IN_THREAD,
						    xcb_atom_t selection)
{
	selection_cursor_t::lock cursor_lock{IN_THREAD, *this};

	create_secondary_selection(IN_THREAD, selection, cursor_lock);
}

bool editor_implObj
::create_secondary_selection(ONLY IN_THREAD,
			     xcb_atom_t selection,
			     selection_cursor_t::lock &cursor_lock)
{
	if (!config.update_clipboards)
		return false;

	if (!cursor_lock.cursor)
		return false;

	if (cursor_lock.cursor->compare(cursor) == 0)
		return false; // Nothing selected, really.
	// Remove the previous one.

	remove_secondary_selection(IN_THREAD);

	auto s=ref<secondary_selectionObj>
		::create(get_screen()->impl->thread->timestamp(IN_THREAD),
			 ref(this), cursor_lock.cursor, selection);

	secondary_selection(IN_THREAD)=s;
	get_window_handler().selection_announce(IN_THREAD, selection, s);
	return true;
}

void editor_implObj::remove_primary_selection(ONLY IN_THREAD)
{
	if (!config.update_clipboards)
		return;

	{
		mpobj<primary_selectionptr>::lock
			lock{current_primary_selection};

		if (!*lock)
			return;

		(*lock)->stillvalid(IN_THREAD)=false;
		*lock=nullptr;
	}

	get_window_handler().selection_discard(IN_THREAD, XCB_ATOM_PRIMARY);
}

void editor_implObj::remove_secondary_selection(ONLY IN_THREAD)
{
	if (!config.update_clipboards)
		return;

	if (!secondary_selection(IN_THREAD))
		return;

	auto selection=secondary_selection(IN_THREAD)->selection;

	secondary_selection(IN_THREAD)->stillvalid(IN_THREAD)=false;
	secondary_selection(IN_THREAD)=nullptr;

	get_window_handler().selection_discard(IN_THREAD, selection);
}

bool editor_implObj::to_begin(ONLY IN_THREAD, const input_mask &mask,
				  const callback_trigger_t &trigger)
{
	bool moved;
	{
		moving_cursor moving{IN_THREAD, *this, mask, moved, trigger};

		cursor->swap(cursor->begin());
	}
	return moved;
}

bool editor_implObj::to_end(ONLY IN_THREAD, const input_mask &mask,
				const callback_trigger_t &trigger)
{
	bool moved;

	{
		moving_cursor moving{IN_THREAD, *this, mask, moved, trigger};

		cursor->swap(cursor->end());
	}
	return moved;
}

void editor_implObj::select_all(ONLY IN_THREAD,
				    const callback_trigger_t &trigger)
{
	input_mask mask;

	clear_bidi_override(IN_THREAD);

	to_begin(IN_THREAD, mask, trigger);

	mask.shift=true;

	to_end(IN_THREAD, mask, trigger);
}

void editor_implObj::delete_char_or_selection(ONLY IN_THREAD,
						  const input_mask &mask,
						  const callback_trigger_t
						  &trigger)
{
	modifying_text modifying{IN_THREAD, *this,
				 input_change_type::deleted, trigger};
	delete_selection_info del_info{IN_THREAD, *this, modifying};

	if (del_info.had_selection)
	{
		if (mask.shift)
			create_secondary_selection
				(IN_THREAD,
				 secondary_clipboard(IN_THREAD),
				 del_info.cursor_lock);

		update_content(IN_THREAD,
			       modifying,
			       del_info,
			       U"");
		return;
	}

	delete_char(IN_THREAD, modifying);
}

void editor_implObj::delete_char(ONLY IN_THREAD,
				     modifying_text &modifying)
{
	delete_char(IN_THREAD, modifying, cursor);
}

void editor_implObj::delete_char(ONLY IN_THREAD,
				     modifying_text &modifying,
				     const richtextiterator &delete_cursor)
{
	auto clone=delete_cursor->clone();
	clone->move_for_delete(IN_THREAD);

	auto p=delete_cursor->pos();

	if (p == clone->pos())
		return;

	update_content(IN_THREAD, modifying, {clone, delete_cursor}, U"");
}

std::u32string editor_implObj::get(const std::optional<bidi_format>
				       &embedding)
{
	return get_content(cursor->begin(), cursor->end(),
			   embedding).get_string();
}

size_t editor_implObj::size() const
{
	return cursor->my_richtext->read_only_lock
		([]
		 (const auto &impl)
		 {
			 return (*impl)->num_chars-1;
		 });
}

std::tuple<richtextiterator, richtextiterator> editor_implObj::pos()
{
	selection_cursor_t::const_lock cursor_lock{*this};

	return pos(cursor_lock);
}

std::tuple<richtextiterator, richtextiterator>
editor_implObj::pos(selection_cursor_t::const_lock &cursor_lock)
{
	auto cursor_pos=cursor_lock.cursor_pos();

	return {cursor, cursor_pos ? richtextiterator{cursor_pos}:cursor};
}

void editor_implObj::set(const std::u32string_view &str, bool validated)
{
	get_window_handler().thread()->run_as
		([str=std::u32string{str}, me=ref{this}, validated]
		 (ONLY IN_THREAD)
		 {
			 me->lock_and_set(IN_THREAD, str, validated);
		 });
}

void editor_implObj::lock_and_set(ONLY IN_THREAD,
				      const std::u32string_view &string,
				      bool validated)
{
	// This invokes the on_change() callback. Editable
	// combo-box's on_change() callback calls
	// the layout manager callback if the input field
	// gets set(), so we should lock the editor element
	// first, here.

	internal_richtext_impl_t::lock lock{text->impl};

	set(IN_THREAD, std::u32string{string}, validated, user_mod{});

}

void editor_implObj::set(ONLY IN_THREAD, const std::u32string &string,
			     bool validated,
			     const callback_trigger_t &trigger)
{
	set(IN_THREAD, string, string.size(), string.size(), trigger);

	// draw_changes() was called, setting validation_required=true
	//
	// Reset it to the app's request.

	validation_required(IN_THREAD)=!validated;
}

void editor_implObj::set(ONLY IN_THREAD, const std::u32string &string,
			     size_t cursor_pos, size_t selection_pos,
			     const callback_trigger_t &trigger)
{
	size_t s=string.size();

	if (s > config.maximum_size)
		return;

	if (cursor_pos > s)
		cursor_pos=s;
	if (selection_pos > s)
		selection_pos=s;

	bool will_have_selection=cursor_pos != selection_pos;

	bool ignored;

	moving_cursor moving{IN_THREAD, *this,
		will_have_selection ? move_type::selection_in_progress
		: move_type::other,
		ignored, trigger
	};
	selection_cursor_t::lock &cursor_lock=moving.cursor_lock;

	update_content(IN_THREAD, moving, { text->begin(),
					    text->end() }, string);
	remove_primary_selection(IN_THREAD);

	cursor->swap(cursor->pos(cursor_pos));

	if (will_have_selection)
	{
		cursor_lock.cursor=cursor->pos(selection_pos);
	}
	else
	{
		cursor_lock.cursor=nullptr;
	}
}

bool editor_implObj::ok_to_lose_focus(ONLY IN_THREAD,
					  const callback_trigger_t &trigger)
{
	return validate_modified(IN_THREAD, trigger);
}


bool editor_implObj::validate_modified(ONLY IN_THREAD,
					   const callback_trigger_t &trigger)
{
	if (!enabled(IN_THREAD, enabled_for::input_focus))
		// We could be here because we're losing keyboard focus after
		// we become invisible. Don't want to invoke validation in
		// that case.
		return true;

	if (!validation_required(IN_THREAD))
		// That was easy, it's already been validated, presumably.
		return true;

	if (!validation_callback(IN_THREAD))
	{
		validation_required(IN_THREAD)=false;
		return true; // By default.
	}

	bool flag=false;

	try {
		auto me=ref{this};

		input_lock lock{&me};

		flag=validation_callback(IN_THREAD)(IN_THREAD, lock, trigger);
	} REPORT_EXCEPTIONS(this);

	if (flag)
		validation_required(IN_THREAD)=false;
	return flag;
}

void editor_implObj::show_droppable_pointer(ONLY IN_THREAD)
{
	set_cursor_pointer(IN_THREAD,
			   cursor_pointer_1tag<dragging_pointer>
			   ::tagged_cursor_pointer(IN_THREAD));
}

void editor_implObj::show_notdroppable_pointer(ONLY IN_THREAD)
{
	set_cursor_pointer(IN_THREAD,
			   cursor_pointer_1tag<dragging_wontdrop_pointer>
			   ::tagged_cursor_pointer(IN_THREAD));
}

bool editor_implObj::show_hint(ONLY IN_THREAD)
{
	if (!hint)
		return false;

	// An empty field has one character, the trailing space.
	if (text->size(IN_THREAD) > 1)
		return false;

	return !current_keyboard_focus(IN_THREAD);
}

bool editor_implObj::should_redraw_to_show_hint(ONLY IN_THREAD)
{
	auto now=show_hint(IN_THREAD);

	if (now == is_showing_hint)
		return false;

	is_showing_hint=now;

	return true;
}

void editor_implObj::on_search(ONLY IN_THREAD,
				   const input_field_config::search_info &)
{
	stop_message(_("Internal error: on_search() is called without enabling"
		       " searching"));
}

LIBCXXW_NAMESPACE_END
